"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import gzip
from os import path
from typing import List, Optional, Tuple

import torch
from torch.nn import Module
from torch.utils.data import TensorDataset, random_split
from tqdm import tqdm

from neural_semigroups.constants import CAYLEY_DATABASE_PATH, CURRENT_DEVICE
from neural_semigroups.denoising_autoencoder import MagmaDAE
from neural_semigroups.magma import Magma
from neural_semigroups.utils import (
    check_filename,
    check_smallsemi_filename,
    download_smallsemi_data,
    get_equivalent_magmas,
    import_smallsemi_format,
)


class CayleyDatabase:
    """
    a database of Cayley tables with different utility functions
    """

    _model: Optional[Module] = None

    def __init__(
        self,
        cardinality: int,
        database_filename: Optional[str] = None,
        data_path: str = CAYLEY_DATABASE_PATH,
    ):
        """
        :param cardinality: the number of elements in underlying magmas
        :param database_filename: a full path to a pre-generated Cayley database.
                                  If ``None``, a ``smallsemi`` data is used.
        :param data_path: a valid path to use as a permanent data storage
        """
        self.cardinality = cardinality
        self.data_path = data_path
        if database_filename is None:
            filename = path.join(
                self.data_path, "smallsemi_data", f"data{cardinality}.gl.gz"
            )
            check_smallsemi_filename(filename)
            if not path.exists(filename):
                download_smallsemi_data(self.data_path)
            with gzip.open(filename, "rb") as file:
                self.database = import_smallsemi_format(file.readlines()).to(
                    CURRENT_DEVICE
                )
            self.labels = torch.ones(
                len(self.database), dtype=torch.int64, device=CURRENT_DEVICE
            )
        else:
            check_filename(path.basename(database_filename))
            torch_zip_file = torch.load(database_filename)
            self.database = torch_zip_file["database"]
            self.labels = torch_zip_file.get(
                "labels", torch.zeros(len(self.database), dtype=torch.int64)
            )

    def augment_by_equivalent_tables(self) -> None:
        """
        for every Cayley table in a previously loaded database adds all of its
        equivalent tables to the database
        """
        database: List[torch.Tensor] = []
        for table in tqdm(
            self.database, desc="augmenting by equivalent tables"
        ):
            database += [get_equivalent_magmas(table)]
        self.database = torch.unique(torch.cat(database, dim=0), dim=0)

    def _check_input(self, cayley_table: List[List[int]]) -> bool:
        """
        checks the input to be a correct Cayley table

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: whether the input is correct
        """
        correct = True
        if isinstance(cayley_table, list):
            # pylint: disable=not-callable
            table = torch.tensor(cayley_table)
            if table.shape != torch.Size([self.cardinality, self.cardinality]):
                correct = False
            elif table.dtype != torch.long:
                correct = False
            elif table.max() >= self.cardinality or table.min() < -1:
                correct = False
        else:
            correct = False
        return correct

    def search_database(
        self, cayley_table: List[List[int]]
    ) -> List[torch.Tensor]:
        """
        get a list of possible completions of a partially filled Cayley table
        (unknown entries are filled by ``-1``)

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: a list of Cayley tables
        """
        if not self._check_input(cayley_table):
            raise ValueError(
                f"invalid Cayley table of {self.cardinality} elements"
            )
        completions = list()
        if self.database is not None:
            # pylint: disable=not-callable
            partial_table = torch.tensor(cayley_table)
            rows, cols = torch.where(partial_table != -1)
            for table in tqdm(
                self.database, desc="full scan over Cayley database"
            ):
                if torch.allclose(
                    table[rows, cols], partial_table[rows, cols]
                ):
                    completions.append(table)
        return completions

    def fill_in_with_model(
        self, cayley_table: List[List[int]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        get a list of possible completions of a partially filled Cayley table
        (unknow entries are filled by ``-1``) using a machine learning model

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: a tuple: (most probable completion, probabilistic cube)
        """
        if isinstance(self.model, MagmaDAE):
            self.model.apply_corruption = False
        if isinstance(self.model, Module):
            self.model.eval()
        if not self._check_input(cayley_table):
            raise ValueError(
                f"invalid Cayley table of {self.cardinality} elements"
            )
        # pylint: disable=not-callable
        table = torch.tensor(cayley_table)
        inv_cardinality = 1 / self.cardinality
        cube = torch.zeros(
            [self.cardinality, self.cardinality, self.cardinality],
            dtype=torch.float32,
            device=CURRENT_DEVICE,
        )
        rows, cols = torch.where(table != -1)
        cube[rows, cols, table[rows, cols]] = 1.0
        rows, cols = torch.where(table == -1)
        cube[rows, cols, :] = inv_cardinality
        prediction = self.model(
            cube.reshape(
                [-1, self.cardinality, self.cardinality, self.cardinality]
            )
        ).detach()[0]
        return (prediction.argmax(axis=-1), prediction)

    def load_model(self, filename: str) -> None:
        """
        load pre-trained PyTorch model

        :param filename: where to load the model from
        :returns:
        """
        self._model = torch.load(filename)

    def train_test_split(
        self, train_size: int, validation_size: int
    ) -> Tuple["CayleyDatabase", "CayleyDatabase", "CayleyDatabase"]:
        """
        split a database of Cayley table in three: train, validation, and test

        :param cayley_db: a database of Cayley tables
        :param train_size: number of tables in a train set
        :param validation_size: number of tables in a validation set
        :returns: a triple of distinct Cayley tables databases: ``(train, validation, test)``
        """
        splits = random_split(
            TensorDataset(self.database, self.labels),
            [
                train_size,
                validation_size,
                self.database.shape[0] - train_size - validation_size,
            ],
        )
        train = CayleyDatabase(self.cardinality)
        train.database, train.labels = [
            torch.stack(i) for i in zip(*splits[0])  # type: ignore
        ]
        validation = CayleyDatabase(self.cardinality)
        validation.database, validation.labels = [
            torch.stack(i) for i in zip(*splits[1])  # type: ignore
        ]
        test = CayleyDatabase(self.cardinality)
        test.database, test.labels = [
            torch.stack(i) for i in zip(*splits[2])  # type: ignore
        ]
        return train, validation, test

    @property
    def model(self) -> Module:
        """
        :returns: pre-trained Torch model
        """
        if self._model is None:
            raise ValueError("The model should be loaded first!")
        return self._model

    @model.setter
    def model(self, model: Module) -> None:
        """
        :param model: pre-trained Torch model
        """
        self._model = model

    def testing_report(self, max_level: int = -1) -> torch.Tensor:
        """
        this function:

        * takes 1000 random Cayley tables from the database
          (if there are less tables, it simply takes all of them)
        * for each Cayley table generates ``max_level`` puzzles
        * each puzzle is created from a table by omitting several cell values
        * for each table the function omits 1, 2, and up to ``max_level`` of all cells
        * each puzzle is given to a pre-trained model of that database
        * if the model returns an associative table
          (not necessary the original one)
          it is considered to be a sucessfull solution
        * in addition, all correctly filled cells are counted
          (despite leading to a full associative table)

        :param max_level: up to how many cells to omit when creating a puzzle;
            when not provided or explicitly set to ``-1`` it defaults to the total number of cells in a table
        :returns: statistics of solved puzzles splitted by the levels of difficulty
                  (number of cells omitted)
        """
        cardinality = self.cardinality
        max_level = cardinality ** 2 if max_level == -1 else max_level
        totals = torch.zeros((3, max_level))
        database_size = len(self.database)
        test_indices = torch.randperm(database_size)[
            : min(database_size, 1000)
        ]
        for i in tqdm(test_indices, desc="generating and solving puzzles"):
            cayley_table = self.database[i]
            for level in range(1, max_level + 1):
                rows, cols = zip(
                    *[
                        (point // cardinality, point % cardinality)
                        for point in torch.randint(
                            0, cardinality ** 2, [level]
                        )
                    ]
                )
                puzzle = cayley_table.clone().detach()
                puzzle[rows, cols] = -1
                solution, _ = self.fill_in_with_model(puzzle.tolist())
                totals[0, level - 1] += 1
                guessed_cells = sum(
                    solution[rows, cols] == cayley_table[rows, cols]
                )
                if Magma(solution).is_associative:
                    guessed_cells = level
                if guessed_cells == level:
                    totals[1, level - 1] += 1
                totals[2, level - 1] += guessed_cells
        return totals



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neural_semigroups']

package_data = \
{'': ['*']}

install_requires = \
['pandas', 'pytorch-ignite', 'tensorboard', 'torch', 'tqdm']

setup_kwargs = {
    'name': 'neural-semigroups',
    'version': '0.3.2',
    'description': 'Neural networks powered research of semigroups',
    'long_description': "[![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/inpefess/neural-semigroups/blob/master/examples/train_a_model.ipynb) [![PyPI version](https://badge.fury.io/py/neural-semigroups.svg)](https://badge.fury.io/py/neural-semigroups) [![CircleCI](https://circleci.com/gh/inpefess/neural-semigroups.svg?style=svg)](https://circleci.com/gh/inpefess/neural-semigroups) [![Documentation Status](https://readthedocs.org/projects/neural-semigroups/badge/?version=latest)](https://neural-semigroups.readthedocs.io/en/latest/?badge=latest) [![codecov](https://codecov.io/gh/inpefess/neural-semigroups/branch/master/graph/badge.svg)](https://codecov.io/gh/inpefess/neural-semigroups)\n\n# Neural Semigroups\n\nHere we try to model Cayley tables of semigroups using neural networks.\n\nThe simplest way to get started is to [use Google Colaboratory](https://colab.research.google.com/github/inpefess/neural-semigroups/blob/master/examples/train_a_model.ipynb)\n\nMore documentation can be found \n[here](https://neural-semigroups.readthedocs.io).\n\n## Motivation\n\nThis work was inspired by [a sudoku\nsolver](https://github.com/Kyubyong/sudoku). A solved Sudoku puzzle\nis nothing more than a Cayley table of a quasigroup from 9 items with\nsome well-known additional properties. So, one can imagine a puzzle\nmade from a Cayley table of any other magma, e. g. a semigroup, by\nhiding part of its cells.\n\nThere are two major differences between sudoku and puzzles based on\nsemigroups:\n\n1) it's easy to take a glance on a table to understand whether it is\na sudoku or not. That's why it was possible to encode numbers in a\ntable cells as colour intensities. Sudoku is a picture, and a\nsemigroup is not. It's difficult to check a Cayley table's\nassociativity with a naked eye;\n\n2) sudoku puzzles are solved by humans for fun and thus catalogued.\nWhen solving a sudoku one knows for sure that there is a unique\nsolution. On the contrary, nobody guesses values in a partially\nfilled Cayley table of a semigroup as a form of amuzement. As a\nresult, one can create a puzzle from a full Cayley table of a\nsemigroup but there may be many distinct solutions.\n",
    'author': 'Boris Shminke',
    'author_email': 'boris@shminke.ml',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inpefess/neural-semigroups',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from collections import OrderedDict
from typing import List, no_type_check

import torch
from torch import Tensor
from torch.nn import BatchNorm1d, Linear, Module, ReLU, Sequential, Softmax2d
from torch.nn.functional import dropout2d

from neural_semigroups.constants import CURRENT_DEVICE


class MagmaDAE(Module):
    """
    Denoising Autoencoder for probability Cayley cubes of magmas
    """

    apply_corruption: bool = True
    do_sampling: bool = False

    def __init__(
        self, cardinality: int, hidden_dims: List[int], corruption_rate: float
    ):
        """
        :param cardinality: the number of elements in a magma
        :param hidden_dims: a list of sizes of hidden layers of the encoder and the decoder
        :param corruption_rate: what percentage of cells from the Cayley table to substitute with uniform random variables
        """
        super().__init__()
        self.cardinality = cardinality
        self.corruption_rate = corruption_rate
        self.input_dim = cardinality ** 3
        encoder_layers: "OrderedDict[str, Module]" = OrderedDict()
        decoder_layers: "OrderedDict[str, Module]" = OrderedDict()
        dims = [self.input_dim] + hidden_dims
        for i in range(len(dims) - 1):
            encoder_layers.update(
                {f"linear0{i}": Linear(dims[i], dims[i + 1], bias=True)}
            )
            encoder_layers.update({f"relu0{i}": ReLU()})
            encoder_layers.update({f"bn0{i}": BatchNorm1d(dims[i + 1])})
            decoder_layers.update({f"bn1{i}": BatchNorm1d(dims[i])})
            decoder_layers.update({f"relu1{i}": ReLU()})
            decoder_layers.update(
                {f"linear1{i}": Linear(dims[i + 1], dims[i], bias=True)}
            )
        self.decoder_layers = Sequential(
            OrderedDict(reversed(decoder_layers.items()))
        ).to(CURRENT_DEVICE)
        self.encoder_layers = Sequential(encoder_layers).to(CURRENT_DEVICE)
        # pylint: disable=not-callable
        self._nearly_zero = torch.tensor([1e-6], device=CURRENT_DEVICE)
        # pylint: disable=not-callable
        self._nearly_one = torch.tensor(
            [1 - (self.cardinality - 1) * 1e-6], device=CURRENT_DEVICE
        )

    def encode(self, corrupted_input: Tensor) -> Tensor:
        """
        represent input cube as an embedding vector

        :param corrupted_input: a tensor with two indices
        :returns: some tensor with two indices and non-negative values
        """
        return self.encoder_layers(corrupted_input.view(-1, self.input_dim))

    def corrupt_input(self, cayley_cube: Tensor) -> Tensor:
        """
        changes several cells in a Cayley table with uniformly distributed
        random variables

        :param cayley_cube: representation of a Cayley table probability distribution
        :returns: distorted Cayley cube
        """
        dropout_norm = (
            1 - self.corruption_rate if self.apply_corruption else 1.0
        )
        return (
            dropout_norm
            * dropout2d(
                cayley_cube.view(
                    -1,
                    self.cardinality * self.cardinality,
                    self.cardinality,
                    1,
                )
                - 1 / self.cardinality,
                self.corruption_rate,
                self.apply_corruption,
            )
            + 1 / self.cardinality
        ).view(-1, self.cardinality, self.cardinality, self.cardinality)

    def decode(self, encoded_input: Tensor) -> Tensor:
        """
        represent an embedding vector as something with size aligned with the
        input

        :param encoded_input: an embedding vector
        :returns: a vector of values from ``0`` to ``1`` (kind of probabilities)
        """
        return (
            Softmax2d()(
                self.decoder_layers(encoded_input)
                .view(-1, self.cardinality, self.cardinality, self.cardinality)
                .transpose(1, 3)
                .transpose(2, 3)
            )
            .transpose(2, 3)
            .transpose(1, 3)
        )

    def reparameterize(self, mu_and_sigma: Tensor) -> Tensor:
        """
        do a reparametrization trick

        :param mu_and_sigma: vector of expectation and standard deviation
        :returns: sample from a distribution
        """
        if self.do_sampling:
            dim = mu_and_sigma.shape[1] // 2
            sample = torch.normal(mu_and_sigma[:, :dim], mu_and_sigma[:, dim:])
        else:
            sample = mu_and_sigma
        return sample

    # pylint: disable=arguments-differ
    @no_type_check
    def forward(self, cayley_cube: Tensor) -> Tensor:
        """
        forward pass inhereted from Module

        :param cayley_cube: probabilistic representation of a magma
        :returns: autoencoded probabilistic representation of a magma
        """
        corrupted_input = self.corrupt_input(cayley_cube)
        encoded_input = self.encode(corrupted_input)
        reparameterized_input = self.reparameterize(encoded_input)
        decoded_input = self.decode(reparameterized_input)
        return torch.where(
            corrupted_input == 1.0,
            self._nearly_one,
            torch.where(
                corrupted_input == 0.0, self._nearly_zero, decoded_input
            ),
        )
